# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from os import remove
from re import T
from threading import local
from typing import *
import os
import sys
import traceback
import logging

from click import Context
import click
from click.termui import style
from click_anno import click_app
from click_anno.types import flag
import click_log

from .errors import GLinkError, ConflictError
from .core import (
    SyncWays, ConflictPolicies,
    add_link, sync_one, get_all_link_ids, list_, remove_link, push_new_gist
)
from .utils import parse_gist_url

class _CliLoggerHandler(click_log.ClickHandler):
    def emit(self, record):
        super().emit(record)
        if record.levelno == logging.ERROR:
            click.get_current_context().abort()

@click_app
class App:
    def __init__(self, debug: flag=False) -> None:
        # setup logger
        logger = logging.getLogger('glink')
        handler = _CliLoggerHandler()
        handler.formatter = click_log.ColorFormatter()
        logger.handlers = [handler]
        if debug:
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)
        self._logger = logger

    def list_(self):
        '''
        list all links.
        '''

        for link_id, link_data in list_().items():
            self._logger.info('{link_id}: {remote_name} {way} {local_name}'.format(
                link_id=click.style(link_id, fg='blue'),
                remote_name='{prov}("{repo}")'.format(
                    prov=link_data['prov'],
                    repo=click.style(link_data['repo'], fg='green')
                ),
                way=SyncWays(link_data['way']).to_symbol(),
                local_name='local("{local_file}")'.format(
                    local_file=click.style(link_data['local_file'], fg='green')
                ),
            ))

    def link(self, ctx: Context, url: str, file: str=None, *, way: SyncWays=SyncWays.twoway):
        'add doc'
        gist_info = parse_gist_url(url)
        if not gist_info:
            return ctx.fail(f'{url} is not a gist url.')

        try:
            link_id = add_link(gist_info, file, way)
            self._logger.info('link id: {}'.format(style(link_id, fg='green')))
            if click.confirm('sync now?', default=True, show_default=True):
                sync_one(link_id)
        except GLinkError as ge:
            ctx.fail(ge.message)

    def unlink(self, link_id: str):
        '''
        remove a link.
        '''
        try:
            remove_link(link_id)
        except KeyError:
            self._logger.error(f'no such link: {link_id}.')
        else:
            self._logger.info(f'unlinked: {link_id}.')

    def push(self, ctx: Context, file: str, user: str=None, public: flag=False):
        'push a file to a new gist'
        if not os.path.isfile(file):
            self._logger.error(f'{file} is not a file.')
        link_id = push_new_gist(file, user=user, public=public)
        self._logger.info('link id: {}'.format(style(link_id, fg='green')))

    def sync(self, link_id: str):
        'sync one link.'

        try:
            sync_one(link_id)
        except ConflictError as e:
            self._logger.warning(e)
        else:
            return

        options = {
            str(ConflictPolicies.local): ConflictPolicies.local,
            str(ConflictPolicies.remote): ConflictPolicies.remote,
            str(ConflictPolicies.skip): ConflictPolicies.skip,
            'unlink': 'unlink'
        }

        choice = click.Choice(options)
        policy = click.prompt('decide to?', type=choice, show_choices=True)

        if policy == 'unlink':
            remove_link(link_id)
        elif policy == str(ConflictPolicies.skip):
            pass
        else:
            sync_one(link_id, options[policy])

    def sync_all(self):
        'sync all links.'

        for link_id in get_all_link_ids():
            self.sync(link_id)

def main(argv=None):
    if argv is None:
        argv = sys.argv
    try:
        App()
    except Exception:
        traceback.print_exc()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import *
import re
import hashlib

def parse_gist_url(url: str):
    match = re.match(r'^(?:https://gist.github.com/(?:(?P<user>[^/]+)/)?)?(?P<gist_id>[0-9a-f]+)(?:#(?P<file>.*))?$', url)
    if match:
        user = match.group('user')
        file = match.group('file')
        rv = dict(
            gist_id=match.group('gist_id'),
        )
        if user:
            rv['user'] = user
        if file:
            rv['file'] = file
        return rv

_GIST_URL_TRANSLATES = {
    ord('.'): '-'
}

def determine_gist_file(expect_name: str, remote_files: List[str]):
    if expect_name in remote_files:
        return expect_name
    if expect_name.startswith('file-'):
        # file-* like gist link
        remote_files_map = dict((x.translate(_GIST_URL_TRANSLATES), x) for x in remote_files)
        remote_file = remote_files_map.get(expect_name[5:])
        if remote_file:
            return remote_file

def sha1_bytes(buffer: bytes):
    return hashlib.sha1(buffer).hexdigest()



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import *
from abc import ABC, abstractmethod
from functools import cache

import requests

class IRemoteProvider(ABC):
    name: str

    @abstractmethod
    def get_remote_version(self, *, user: str, repo: str, remote_file: str,
                           access_token: Optional[str],
                           **kwargs) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_remote_file_content(self, *, user: str, repo: str, remote_file: str,
                                access_token: Optional[str],
                                **kwargs) -> Optional[bytes]:
        '''
        get file content as bytes.
        return `None` if the file was removed.
        '''
        raise NotImplementedError

    @abstractmethod
    def push_local_file_content(self, *, user: str, repo: str, remote_file: str,
                                access_token: Optional[str],
                                local_file_content: bytes,
                                **kwargs) -> str:
        '''
        return the new version string.
        '''
        raise NotImplementedError

    @staticmethod
    @cache
    def _http_get(url: str) -> dict:
        return requests.get(url, timeout=10)



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------


class GLinkError(Exception):
    'the base error for glink'

    def __init__(self, message: str) -> None:
        self.message = message


class LocalFileRemovedError(GLinkError):
    pass


class RemoteFileRemovedError(GLinkError):
    pass


class ConflictError(GLinkError):
    pass



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

from typing import *
from enum import IntEnum, auto
import os
import pathlib
import logging

import requests
from click import style
from typeguard import typechecked
import github

from .errors import GLinkError, LocalFileRemovedError, ConflictError, RemoteFileRemovedError
from .abc import IRemoteProvider
from .configs import GLinkConfigs
from .provs.gist import GistProvider
from .utils import determine_gist_file, sha1_bytes


class SyncWays(IntEnum):
    pull = 1
    push = 2
    twoway = 3

    def __str__(self) -> str:
        return self.name

    def __format__(self, format_spec: str) -> str:
        return self.name

    def to_symbol(self) -> str:
        if self == SyncWays.twoway:
            return '<->'
        elif self == SyncWays.pull:
            return ' ->'
        else:
            return '<- '


class ConflictPolicies(IntEnum):
    unset = auto()
    local = auto()
    remote = auto()
    skip = auto()

    def __str__(self) -> str:
        return self.name

    def __format__(self, format_spec: str) -> str:
        return self.name


_CONFIGS = GLinkConfigs()

def add_link(gist_info: dict, local_file: Optional[str], way: SyncWays):
    gist_id: str = gist_info['gist_id']
    file: Optional[str] = gist_info.get('file')

    gist_url = f'https://api.github.com/gists/{gist_id}'
    r = requests.get(gist_url, timeout=10)
    gist_data = r.json()

    # determine remote file
    remote_files: List[str] = list(gist_data['files'])

    if file:
        remote_file = determine_gist_file(file, remote_files)
        if not remote_file:
            raise GLinkError(f'no file named "{file}" in gist {gist_id}.')
    else:
        if len(remote_files) == 1:
            remote_file = remote_files[0]
        else:
            raise GLinkError(f'must select a determinate file.')

    if not local_file:
        local_file = os.path.basename(remote_file)
    local_file = os.path.abspath(local_file)

    link_id: str = _CONFIGS.add_link(
        prov='gist',
        user=gist_data['owner']['login'],
        repo=gist_id,
        remote_file=remote_file,
        local_file=local_file,
        way=way.value
    )

    glink_logger.info(f'link added: gist/{gist_id}/{remote_file} {way.to_symbol()} {local_file}')
    return link_id

def push_new_gist(local_file: str, user: Optional[str], public: bool):
    prov = 'gist'

    if user is None:
        users = _CONFIGS.get_users(prov)
        if len(users) == 0:
            glink_logger.error('no gist user in configs.')
        elif len(users) > 1:
            glink_logger.error('you must type a user.')
        else:
            user = users[0]

    auth_info = _CONFIGS.read_auth_info(prov, user)
    if isinstance(auth_info, str):
        access_token = auth_info
    else:
        access_token = None

    provider = GistProvider()
    remote_file = os.path.basename(local_file)
    repo = provider.new_gist(
        user=user,
        filename=remote_file,
        content=pathlib.Path(local_file).read_bytes(),
        access_token=access_token,
        public=public
    )

    link_id: str = _CONFIGS.add_link(
        prov=prov,
        user=user,
        repo=repo,
        remote_file=remote_file,
        local_file=os.path.abspath(local_file),
        way=SyncWays.twoway.value
    )

    glink_logger.info(f'link added: gist/{repo}/{remote_file} {SyncWays.twoway.to_symbol()} {local_file}')
    return link_id

@typechecked
def _sync_one_core(
        prov: str, user: str, repo: str, remote_file: str, local_file: str, way: int, sync_state: dict,
        conflict_policy: ConflictPolicies
    ):
    assert way in SyncWays.__members__.values()

    provider: IRemoteProvider
    if prov == 'gist':
        provider = GistProvider()
    else:
        raise NotImplementedError
    remote_name = f'{prov}("{repo}")'
    local_name = f'local("{local_file}")'

    kwargs = dict(prov=prov, user=user, repo=repo, remote_file=remote_file)
    auth_info = _CONFIGS.read_auth_info(prov, user)
    if isinstance(auth_info, str):
        kwargs['access_token'] = auth_info

    remote_version = provider.get_remote_version(**kwargs)
    if remote_version != sync_state.get('remote_version'):
        glink_logger.debug(f'found new remote version: {remote_version}.')
        remote_file_content = provider.get_remote_file_content(**kwargs)
        if remote_file_content is None:
            glink_logger.info(f'remote file "{remote_file}" is removed, sync is skiped.')
            return
        remote_file_sha1 = sha1_bytes(remote_file_content)
        remote_file_changed = remote_file_sha1 != sync_state.get('file_sha1')
    else:
        glink_logger.debug(f'remote version is not changed: {remote_version}.')
        remote_file_sha1 = None
        remote_file_content = None
        remote_file_changed = False

    local_file_pathobj = pathlib.Path(local_file)
    if os.path.isfile(local_file):
        local_file_content = local_file_pathobj.read_bytes()
        local_file_sha1 = sha1_bytes(local_file_content)
        local_file_changed = local_file_sha1 != sync_state.get('file_sha1')
    elif sync_state:
        raise LocalFileRemovedError(f'local("{local_file}") is removed')
    else:
        local_file_content = None
        local_file_sha1 = None
        local_file_changed = False

    file_sha1 = None
    pull, push = False, False
    if remote_file_changed and local_file_changed:
        if remote_file_sha1 == local_file_sha1:
            glink_logger.info(f'reattach local file "{local_file}" as unchanged.')
            file_sha1 = remote_file_sha1
        else:
            if conflict_policy == ConflictPolicies.unset:
                raise ConflictError(f'{local_name} and {remote_name} both changed.')
            elif conflict_policy == ConflictPolicies.local:
                if way == SyncWays.pull:
                    glink_logger.warning('ignore by pull only.')
                    return
                push = True
            else:
                if way == SyncWays.push:
                    glink_logger.warning('ignore by push only.')
                    return
                pull = True
    elif remote_file_changed:
        if way == SyncWays.push:
            glink_logger.debug('ignore by push only.')
            return
        pull = True
    elif local_file_changed:
        if way == SyncWays.pull:
            glink_logger.debug('ignore by pull only.')
            return
        push = True

    assert not (pull and push)
    if pull:
        local_file_pathobj.write_bytes(remote_file_content)
        file_sha1 = remote_file_sha1
        glink_logger.info(f'pull {remote_name} to {local_name}.')
    elif push:
        remote_version = provider.push_local_file_content(local_file_content=local_file_content, **kwargs)
        file_sha1 = local_file_sha1
        glink_logger.info(f'push {local_name} to {remote_name}.')

    assert remote_version
    assert file_sha1
    sync_state.update(
        remote_version=remote_version,
        file_sha1=file_sha1
    )
    return True

def sync_one(link_id: str, conflict_policy: ConflictPolicies=ConflictPolicies.unset):
    link_data: dict = _CONFIGS.get_link(link_id=link_id)
    sync_state = link_data.setdefault('sync_state', {})
    synced = False
    try:
        synced = _sync_one_core(conflict_policy=conflict_policy, **link_data)
    except (LocalFileRemovedError, RemoteFileRemovedError) as e:
        glink_logger.warning(f'skiped link("{link_id}") because {e.message}.')
    if synced:
        _CONFIGS.save_state(link_id=link_id, sync_state=sync_state)

def get_all_link_ids():
    return _CONFIGS.get_all_link_ids()

def list_():
    return dict(
        (link_id, _CONFIGS.get_link(link_id=link_id)) for link_id in _CONFIGS.get_all_link_ids()
    )

def remove_link(link_id: str):
    _CONFIGS.remove_link(link_id=link_id)


glink_logger = logging.getLogger('glink')



# -*- coding: utf-8 -*-
#
# Copyright (c) 2020~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

import json
from functools import partial
import uuid
import json

import xdg
import sqlitedict
from typeguard import typechecked

class GLinkConfigs:
    def __init__(self) -> None:
        self._conf_root = xdg.XDG_CONFIG_HOME / 'Cologler' / 'glink'
        self._conf_root.mkdir(parents=True, exist_ok=True)
        self._auth_path = self._conf_root / 'auth.json'

    def _get_linkdb(self, autocommit=False):
        return sqlitedict.SqliteDict(
            filename=str(self._conf_root / 'glink.db'),
            tablename='links',
            journal_mode='WAL',
            autocommit=autocommit,
            encode=partial(json.dumps, ensure_ascii=False),
            decode=json.loads
        )

    @typechecked
    def add_link(self, prov: str, user: str, repo: str, remote_file: str, local_file: str, way: int):
        id = str(uuid.uuid4())
        with self._get_linkdb(True) as linkdb:
            linkdb[id] = dict(
                prov=prov,
                user=user,
                repo=repo,
                remote_file=remote_file,
                local_file=local_file,
                way=way
            )
        return id

    @typechecked
    def get_link(self, link_id: str):
        with self._get_linkdb() as linkdb:
            return linkdb.get(link_id)

    @typechecked
    def remove_link(self, link_id: str):
        with self._get_linkdb(True) as linkdb:
            del linkdb[link_id]

    @typechecked
    def save_state(self, link_id: str, sync_state: dict):
        with self._get_linkdb(True) as linkdb:
            data = linkdb[link_id]
            data['sync_state'] = sync_state
            linkdb[link_id] = data

    def get_all_link_ids(self):
        with self._get_linkdb() as linkdb:
            return list(linkdb)

    @typechecked
    def get_users(self, prov: str):
        'get all user for the service'
        if self._auth_path.is_file():
            text = self._auth_path.read_text(encoding='utf-8')
            d: dict = json.loads(text)
            suffix = f'@{prov}'
            return [k.removesuffix(suffix) for k in d if k.endswith(suffix)]
        return []

    @typechecked
    def read_auth_info(self, prov: str, user: str):
        if self._auth_path.is_file():
            text = self._auth_path.read_text(encoding='utf-8')
            return json.loads(text).get(f'{user}@{prov}')
        return None
